<?php
/*
    Copyright (C) 2007 Yan Morin <yan.morin@gmail.com>

    This file is part of libfdpf.

    libfdpf is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    libfdpf is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libfdpf.  If not, see <http://www.gnu.org/licenses/>.
*/

/**
 * php fdpf.php filename [--debug] > generate-pdf.php
 *
 * @arguments
 *   [1] "filename" PDF file to convert to fpdf/svg format
 *   [2] "--debug"  activates debug
 */
//init global var
$filename = '';
$debug = false;

// getopts
if ($_SERVER['argc'] >= 2 && file_exists($_SERVER['argv'][1])) {
	$filename = $_SERVER['argv'][1];
}
if ($_SERVER['argc'] >= 3 && $_SERVER['argv'][2] == '--debug') {
	$debug = true;
}

$dpdf = new DPDF($filename);
//require_once 'fdpf_svgwriter.php';
//$dpdf->setWriter(new FDPF_SVGWriter());

// using fdpf writer
require_once 'fdpf_fpdfwriter.php';
$dpdf->setWriter(new FDPF_FPDFWriter());
$dpdf->Output();

class Type {
	var $propertiesA;
	var $propertiesValueA;
	var $stream;
	var $tokenname;

	function Type($tokenname = '') { 
		$this->tokenname = $tokenname;
		$this->propertiesA = array(); 
		$this->propertiesValueA = array(); 
		$this->propertiesRefA = array();
	}
	function get($name, &$objmap = null) { 
		if (array_key_exists($name, $this->propertiesValueA)) {
			$value = $this->propertiesValueA[$name];
			if ($value === null && $objmap !== null) {
				$value = $objmap[$this->propertiesRefA[$name]];
			}
			return $value;
		}
		return null;
	}
	function load($str, $everything=false) {
		while ( ($name = $this->_getToken($str)) != FALSE ) {
			//_debug('load name: ' . $name . "\n");
			$value = '';
			if ($str == '' || $str{0} == '/') {
				// we get another properties, this is a bool
				$value = true;
			} else {
				// we got a space, so discard it
				if ($str{0} == ' ' ) { $str = substr($str, 1); }
				if ($str == '') {
					$value = true;
				} else {
					$char = $str{0};
					if ($char == '/') {
						$value = $this->_getToken($str);
					} else if ($char == '[') {
						$value = $this->_getArray($str);
					} else if ($char == '(') {
						$value = $this->_getSubstr($str);
					} else if ($char == '<') {
						if ($str{1} == '<') { // find a << >>
							$p = strpos($str, '>>');
							$value = substr($str, 2, $p-2);
							$str = substr($str, $p+2);
						} else {
							$p = strpos($str, '>'); // find a < >
							$value = substr($str, 1, $p-1);
							$str = substr($str, $p+1);
						}
						// strip slashes
						$str = ltrim($str);
					} else {
						$value = $this->_getNumber($str);
						if ($str!='' && $str{0} == '0' && $str{2} == 'R') {
							$this->_moveAfterRef($str);
							$this->propertiesRefA[$name] = $value;
							$value = null;
						}
					}
				}
			}
			$this->propertiesValueA[$name] = $value;
		}
		_debug('Load: ' . $this);
	}
	function __toString() {
		$str = '';
		foreach($this->propertiesValueA as $key=>$value) {
			if ($value === null) {
				$str .= "$key => Ref#{$this->propertiesRefA[$key]}\t";
			} else {
				$str .= "$key => $value\t";
			}
		}
		return 'Type('.$this->tokenname.'): ' . $str .  "\n";
	}

	function _getToken(&$str) {
		//_debug("Type::_getToken: [$str]\n");
		$str = ltrim($str);
		$strlen = strlen($str);
		if ($strlen > 2 && $str{0} == '/') {
			$token = '';
			for ($i = 1; $i<$strlen; $i++) {
				$char = $str{$i};
				if ($char == '/' || $char == ' ' || $char == '[' || $char == '<' || $char == '(') {
					$str = substr($str, strlen($token) + 1); // string contain the space
					return $token;
				} else {
					$token .= $char;
				}
			}
		}
		return FALSE;
	}
	function _getRef(&$str) {
		$idx = '';
		if ($str != '') {
			$p = strpos($str, ' ');
			if ($p !== FALSE) {
				$idx = substr($str, 0, $p);
				$str = substr($str, $p+5);
			}
		}
		return $idx;
	}
	function _moveAfterRef(&$str) {
		if ($str{3} == ' ') {
			$str = substr($str, $p+4);
		} else {
			$str = substr($str, $p+3);
		}
	}
	function _getNumber(&$str) {
		$num = '';
		if ($str != '') {
			$p1 = strpos($str, ' ');
			$p2 = strpos($str, '/');
			if ($p1 !== FALSE || $p2 !== FALSE) {
				$p = ($p1 === FALSE) ? $p2 : (($p2 === FALSE) ? $p1 : ($p1 < $p2) ? $p1 : $p2 );
				$num = substr($str, 0, $p);
				$str = ltrim(substr($str, $p));
			}
		}
		return $num;
	}
	function _getArray(&$str) {	
		$a = array();
		$endpos = strpos($str, ']');
		if ($endpos !== FALSE) {
			$as = trim(substr($str, 1, $endpos-1));
			$a = explode(' ',$as);
			$str = substr($str, $endpos+1);
			if ($str{0} == ' ') { // strip space
				$str = substr($str, 1);
			}
		}
		return $a;
	}
	function _getSubstr(&$str) {
		$substr = '';
		$endpos = strpos($str, ')');
		if ($endpos !== FALSE) {
			$substr = substr($str, 1, $endpos-1);
			$str = substr($str, $endpos+2);
		}
		return $substr;
	}

	function getStream() {
		if ((array_key_exists('Filter', $this->propertiesValueA) && 
			$this->propertiesValueA['Filter'] == 'FlateDecode') || 
		     array_key_exists('FlateDecode', $this->propertiesValueA)) {
			return gzuncompress($this->stream);
		}
		return $this->stream;
	}
}

class Filter extends Type {
	function Filter() {
		parent::Type('Filter');
		$this->propertiesA = array('FlateDecode'=>'Bool','Length'=>'Number');
	}
}

class Font extends Type {
	// /BaseFont /Helvetica-Bold
	// /Subtype /Type1
	// /Encoding /WinAnsiEncoding
	var $family = null;
	var $fonttype = null;
	function Font() { 
		parent::Type('Font');
		$this->propertiesA = array(
			'BaseFont'=>'String',
			'Subtype'=>'String',
			'Encoding'=>'String'
		); 
	}

	function getFamily() {
		if ($this->family === null && array_key_exists('BaseFont', $this->propertiesValueA)) {
			$this->family = $this->get('BaseFont');
			$p = strrpos($this->family,'-');
			if ($p !== FALSE) {
				$last = substr($this->family, $p+1);
				if (stristr($last,'Bold') == 0 || 
				    stristr($last,'BoldMT') == 0) {
					$this->family = substr($this->family, 0, $p);
					$this->fonttype = 'B';
				}
			}
			$first = substr($this->family, 0, 7);
			if ($first == 'CAAAAA+' || $first == 'BAAAAA+' || $first == 'DAAAAA+') {
				$this->family = substr($this->family, 7);
			}
			if ($this->family == 'ArialMT' || $this->family = 'DejaVuSans') {
				$this->family = 'Arial';
			}
		}
		return $this->family;
	}
	function getFontType() {
		if ($this->fonttype === null) $this->getFamily();
		return $this->fonttype;
	}
}

class Resources extends Type {
	var $ProcSet;
	var $Font;

	function Resources($procset) {
		parent::Type('Resources');
		$this->ProcSet = $procset;
		$this->propertiesA = array('F1'=>'Ref');
	}
	function HasFont() {
		return (array_key_exists('F1', $this->propertiesValueA)) ? true : false;
	}
}

class Pages extends Type {
	var $kids;
	var $count;
	var $mediabox;
	function Pages() {
		parent::Type('Pages');
		$this->propertiesA = array('Kids'=>'Array', 'Count'=>'Number', 'MediaBox'=>'Array');
	}
	
}

class DPDF {
	var $fileA = null;
	var $l = 0;
	var $is_valid = false;
	var $pdfversion = '1.3';
	var $objmap; //idx -> object
	var $trailer = '';
	var $noimage = 0;
	var $writer = null;

	function DPDF($filename) {
		$this->fileA = file($filename);
		if ($this->fileA) {
			$this->objmap = array();
			if ($this->_decode()) {
				$this->is_valid = true;
			}
		}
	}

	function setWriter($newWriter) {
		$this->writer = $newWriter;
	}

	function Output() {
		if ($this->is_valid && $this->trailer!=null) {
		if ($this->writer == null) { exit('You need to set a FDPF Writer()!'); }

		$this->orientation = 'P';
		$this->unit = 'pt';
		$this->format = 'A4';
		$this->pagewidth = 0;  // TODO
		$this->pageheight = 0; // TODO

		// start
		$rootCatalog = $this->trailer->get('Root', $this->objmap);
		if ($rootCatalog != null) {
			$pages = $rootCatalog->get('Pages', $this->objmap);
			$args = '';
			if ($pages != null) {
				$mediabox = $pages->get('MediaBox');
				$format = '';
				if ($mediabox !== null) {
					$this->pagewidth = $w = intval($mediabox[2]);
					$this->pageheight = $h = intval($mediabox[3]);
					if ($w > $h) {
						$orientation = 'L';
						// should we switch coordinates?
						$a = $h; $h = $w; $w = $a;
					}
				}
			}
		}

		$this->writer->header($this->orientation, $this->unit, $this->format, $this->pagewidth, $this->pageheight);
		if ($pages != null) {
			$pagesCount = $pages->get('Count');
			$kids = $pages->get('Kids');
			for ($i = 0; $i < $pagesCount; $i++) {
				$this->_echoPage($kids[$i*3]);
			}
		}
		$this->writer->footer();
		} else {
			echo "Invalid format\n";
		}
	}


	function _nextLine() {
		//echo "LINE {$this->l}: {$this->fileA[$this->l]}";
		return rtrim($this->fileA[$this->l++], "\n\r");
	}

	function _decode() {
		// first line contain %PDF-Version
		$line = $this->_nextLine();
		if (substr($line,0,5) === '%PDF-') {
			$this->pdfversion = substr($line,6);
		} else {
			return false;
		}
		$line = $this->_nextLine();
		while ( $line != '%%EOF') {
			while ($line == '' || $line{0} == '%') {
				if ($line == '%%EOF') {
					return true;
				}
				$line = $this->_nextline(); // %Ã¤Ã¼Ã¶Ã<9f> in open office
			}
			//echo "#{$this->l}\n";
			$typeA = explode(' ', $line);
			if (count($typeA)) {
				switch($typeA[count($typeA)-1]) {
					case 'obj':
						$this->_decodeObj($typeA[0], $typeA[2]);
					break;
					case 'xref':
						$this->_decodeXref();
					break;
					case 'trailer':
						$this->_decodeTrailer();
					break;
					case 'startxref':
						$this->_decodeStartxref();
					break;
					default:
						_debug('_decode Unknown: ' . $typeA[count($typeA)-1]);
				}
			}
			$line = $this->_nextLine();
		}
		return true;
	}
	function _decodeObj($idx, $zero) {
		_debug('_decodeObj idx='.$idx);
		// <</bla>>
		// endobj
		$line = $this->_nextLine();
		// there should be something like 'nextToken() here'
		if (substr($line, 0, 2) == '<<') {
			$line = ltrim(substr($line, 2));
			$list = $this->_decodePropertyBlock($idx, $line);
		} else if ($line != 'endobj') {
			$this->objmap[$idx] = trim($line);
			//_debug("obj #$idx = " . trim($line));
		}
		while($line != 'endobj') {
			$line = $this->_nextLine();
		}
	}

	function _getTokenName(&$str) {
		$str = ltrim($str);
		if ($str=='') {
			$str = ltrim($this->_nextLine());
		}
		if ($str{0} == '/') {
			$end = false;
			$i = 1;
			$len = strlen($str);
			$token = '';
			while(!$end) {
				if ($i < $len) {
					$char = $str{$i};
					if ($char == '/' || $char == '<') {
						$str = substr($str, $i);
						return $token;
					} else if ($char == ' ') {
						$str = substr($str, $i+1);
						return $token;
					} else {
						$token .= $char;
					}
					$i++;
				} else {
					$str = $this->_nextLine();
					$len = strlen($str);
					$i = 0;
				}
			}
		}
		return '';
	}

	function _decodePropertyBlock($idx, &$line) {
		$p = 0;
		$token = $this->_getTokenName($line);
		if ($token == '') {
			$this->objmap[$idx] = '';
			return;
		}
		switch($token) {
			case 'Type':
				$token = $this->_getTokenName($line);
				//echo "Loading /Type $token\n";
				switch($token) {
					case 'Pages':
						$pages = new Pages();
						$str = $this->_getPropertiesString($line);
						$pages->load($str);
						$this->objmap[$idx] = $pages;
					break;
					case 'Font':
						$font = new Font();
						$str = $this->_getPropertiesString($line);
						$str = str_replace('/SubType/','/SubType /', $str);
						$str = str_replace('/BaseFont/','/BaseFont /', $str);
						$font->load($str);
						$this->objmap[$idx] = $font;
					break;
					case 'FontDescriptor':
						$type = new Type($token);
						$str = $this->_getPropertiesString($line);
						$str = str_replace('/FontName/','/FontName /', $str);
						$type->load($str);
						$this->objmap[$idx] = $type;
					break;
					case 'Page':
					case 'Catalog':
						$type = new Type($token);
						$str = $this->_getPropertiesString($line);
						$type->load($str);
						$this->objmap[$idx] = $type;
					break;
					default:
						_debug('Unhandle /Type: ' . $token);
				}
			break;
			case 'Filter':
				$filter = new Filter();
				$string = $this->_getPropertiesString($line);
				$filter->load($string);
				// bug here, we have an attribute d'attributes... /Filter /FlateDecode /Length 71
				// get $filter->get('Number') octets
				if ($line == 'stream') {
					$filter->stream = $this->_getStream($line);
				}
				$this->objmap[$idx] = $filter;
				// 19>>
			break;
			default:
				$type = new Type();
				$line = '/'.$token . ' ' . $line;
				$string = $this->_getPropertiesString($line);
				_debug('obj #' . $idx . ' ');
				$type->load($string, true);
				if ($line == 'stream') {
					$type->stream = $this->_getStream($line);
				}
				$this->objmap[$idx] = $type;
		}
	}
	// from a <</Type aaa [...>>]
	function _getPropertiesString(&$str) {
		$returnstr = '';
		$strlen = strlen($str);
		$endfound = false;
		while (!$endfound) {
			if ($str == '') { $str = $this->_nextLine(); }
			if (($p = strpos($str, '<<')) !== FALSE) {
				$p2 = strpos($str, '>>', $p+2);
				$returnstr .= substr($str, 0, $p2+2);
				$str = substr($str, $p2+2);
			} else if (($p = strpos($str, '>>')) !== FALSE) {
				$returnstr .= substr($str, 0, $p) . ' ';
				$str = $this->_nextLine();
				$endfound = true;
				break;
			} else {
				$returnstr .= $str . ' ';
				$str = $this->_nextLine();
			}
		}
		//_debug("_getPropertiesString: [$returnstr]");
		return $returnstr;
	}

	function _decodeXref() {
		_debug('_decodeXref');
		$line = $this->_nextLine();
		list($no, $nbline) = explode(' ', $line); // 0 7
		for ($i = 0; $i < $nbline; $i++) {
			$line = $this->_nextLine(); // 0000000000 65535 f
			                            // 0000000175 00000 n
		}
	}

	function _decodeTrailer() {
		_debug('_decodeTrailer');
		$line = $this->_nextLine(); // <<
		if (strpos($line,'<<') === 0) {
			$line = substr($line,2);
			$trailer = new Type();
			$trailer->propertiesA = array('Size'=>'Number', 'Root'=>'Ref','Info'=>'Ref');
			$stringProperties = $this->_getPropertiesString($line);
			$trailer->load($stringProperties);
			$this->trailer = $trailer;
		}
	}

	function _decodeStartxref() {
		_debug('_decodeStartxref');
		$line = $this->_nextLine(); // 535
	}

	function _getStream(&$line) {
		$stream = '';
		$line = $this->_nextline();
		while ( $line != 'endstream' && $line != 'endobj') {
			$stream .= $line."\n";
			$line = $this->_nextline();
		}
		if ($line == 'endobj') {
			$stream = substr($stream, 0, -9); // oups, too much...
		} else {
			$line = $this->_nextline();
		}
		return $stream;
	}

	function _echoPage($idxPage) {
		$page = $this->objmap[$idxPage];
		$parent = $page->get('Parent', $this->objmap);
		$mediabox = $page->get('MediaBox');
		// landscape or portrait?
		if ($mediabox === null) {
			$mediabox = $parent->get('MediaBox',$this->objmap); 
		}
		$pageOrientation = $this->orientation;
		$pageheight = $this->pageheight;
		if ($mediabox !== null) {
			$pagewidth  = floatval($mediabox[2]);
			$pageheight = floatval($mediabox[3]);
			if ($pagewidth > $pageheight) {
				$pageOrientation = 'L' ;
				$ph = $pagewidth; $pw = $pageheight; 
			} else {
				$pageOrientation = 'P';
				$pw = $pagewidth; $ph = $pageheight; 
			}
		}
		$this->writer->page_start($pageOrientation, $pw, $ph);
		$resources = $page->get('Resources', $this->objmap);
		if ($resources === null) { // check for parent
			$resources = $parent->get('Resources', $this->objmap);
		}
		$fonts = $resources->get('Font', $this->objmap);
		$contents = $page->get('Contents', $this->objmap);
		$streamLines = explode("\n", $contents->getStream());
		_debug("====================\nPage stream: ". $contents->getStream() . "======================");
		$lastfontsize = 10;

		$mode = '';
		$modetext = false; // sometimes <000AD9707050A> can be broke in two lines
		$modetext_string = '';
		$h = 0;
		$w = 0;
		$positionx0 = '31.19';  // left
		$positiony0 = '813.55'; // bottom
		$modeline = false;
		$previouspoints=array(); // array of points x1, y1, x2, y2, ...

		foreach($streamLines as $line) { // mega pdf stream parsing!
			$argv = explode(' ', $line);
			$argc = count($argv);
			for ($i = 0; $i < $argc; $i++) {
				$arg = $argv[$i]; 
				if ($modetext) { // we have a string to print!
					$p = strpos($arg, '>');
					if ($p !== FALSE) {
						$modetext = false;
						$modetext_string .= substr($arg,0,$p);
						$txt = $this->_translate($font, $modetext_string);
						$this->writer->text($xconvert, $yconvert, $txt);

						$i++; // Tj should be after
						continue; // next token
					} else {
						$modetext_string .= $arg;
						break; // next line
					}
				}
				if ($modeline) {
					// I don't know how to generate x1 y1 m x2 y2 l x3 y3 l x4 y4 l h
					// but it's not a line...
					if ($arg == 'h') { // draw a rect with 4 points
						$rx = $previouspoints[0];
						$ry = $previouspoints[1];
						$rw = $previouspoints[4]-$rx;
						$rh = $previouspoints[5]-$ry;
						//if ($rw < 1) { $rw = ceil($rw); } // another hack
						//if ($rh < 1) { $rh = ceil($rh); }
						$this->writer->rect($rx,$ry,$rw,$rh,$arg);
						$modeline = false;
						$previouspoints = array();
						continue;
					} elseif ($arg == 'S') { // draw line
						for ($i = 0; $i < count($previouspoints)-2; $i+=2) {
							$this->writer->line($previouspoints[$i],$previouspoints[$i+1],$previouspoints[$i+2],$previouspoints[$i+3]);
						}
						$modeline = false;
						$previouspoints = array();
						continue;
					}
				}
				if ($arg == 'ET') { $mode = ''; break; } // end mode
				if ($mode == 'BT') {
					//_debug("Arg #$i: $arg");
					if ($arg{0} == '/' && $argv[$i+2] == 'Tf') { // it's a font?
						$font = $fonts->get(substr($arg, 1), $this->objmap);
						if ($font !== null) {
							$lastfontsize = $argv[$i+1];
							$this->writer->font($font->getFamily(), $font->getFontType(), $lastfontsize);	
						}
						$i+=2; // get over /F1 10.0 Tf
					} else if ($arg{0} == '(') { // string end with )
						$p = strpos(')', $arg);
						if ($p !== FALSE) {
							$txt = substr($arg,0,$p-1);
						} else {
							$txt = substr($arg,1);
							while(++$i < $argc) {
								$word = $argv[$i];
								if (substr($word,-1,1) == ')') {
									$str .= ' ' . substr($word,0,-1);
									break;
								}
								$txt .= ' ' . $word;
							}
						}
						$this->writer->text($xconvert, $yconvert, $txt);
						$i++; // get over <> Tj
					} elseif ($arg{0} == '<') { // string end with >
						$max = strlen($arg)-1;
						if ($arg{$max} == '>') {
							$txt = substr($arg,1,-1);
							$txt = $this->_translate($font, $txt);
							$this->writer->text($xconvert, $yconvert, $txt);
							$i++; // get over <> Tj
						} else {
							$modetext = true;
							$modetext_string = substr($arg,1);
						}
					} elseif ($argv[$i+2] == 'Td') { // table data 
						$x = $arg;
						$y = $argv[$i+1];
						$xconvert = floatval($x);
						$yconvert = floatval($pageheight - $y);
						//$w = intval($positionx0 - $x);
						//$h = intval(($positiony0 - ($lastfontsize * 0.3) - $y)/1.417);
						$i+=2; // get over 0.0 0.0 Td
					}
				} elseif ($i == 0 && $arg == 'BT') {
					$mode = 'BT';
				} else {
				if ($modeline == true && ($arg == 'h' || $arg == 'S')) {
					$modeline = false;
					continue;
				}
				$rest = $argc-$i;
				if ($rest > 1) {
					if ($argv[$i+1] == 'w') { // line width
						$this->writer->linewidth($argv[$i]);
						$i++; // skip the w
						continue;
					}
				}
				if ($rest > 2) {
					if ($argv[$i+2] == 'm') {
						$previouspoints[] = floatval($arg);
						$previouspoints[] = floatval($pageheight - $argv[$i+1]);
						$modeline = true;
						$i+=2; continue; 
					}
					if ($modeline == true && $argv[$i+2] == 'l') {
						$previouspoints[] = floatval($arg);
						$previouspoints[] = floatval($pageheight - $argv[$i+1]);
					}
				}
				if ($rest > 3) {
					if ($argv[$i+3] == 'RG' || $argv[$i+3] == 'rg') {
						$r = ceil(floatval($arg)*255);
						$g = ceil(floatval($argv[$i+1])*255);
						$b = ceil(floatval($argv[$i+2])*255);
						if ($argv[$i+3] == 'rg') {
							$this->writer->fillcolor($r,$g,$b);
						} else {
							$this->writer->drawcolor($r,$g,$b);
						}
						$i+=3;
						continue;
					}
				}
				if ($rest > 5) {
					if ($argv[$i+4] == 're') {
						switch($argv[$i+5]) {
						}
						if ($argv[$i+5] == 'W*') {
							// this is the border around the page
							// don't do it
							$i+=5; continue;
						}
						$rx = floatval($arg);
						$ry = floatval($pageheight - $argv[$i+1]);
						$rw = floatval($argv[$i+2]);
						$rh = floatval(-1*$argv[$i+3]);
						$this->writer->rect($rx,$ry,$rw,$rh,$argv[$i+5]);
						$i+=5; continue;
					}
				}
				if ($rest > 7) {
					if ($arg == 'q' && $argv[$i+7] == 'cm') { // images
						$iw = floatval($argv[$i+1]);
						$ih = floatval($argv[$i+4]);
						$ix = floatval($argv[$i+5]);
						$iy = floatval($pageheight-$argv[$i+6]-$ih);
						$this->writer->image($this->noimage, $ix, $iy, $iw, $ih);
						$this->noimage++;
						$i+=7; continue; // q 20.00 0 0 22.00 100.00 620.00 cm
						// there is the /Im3 Do Q after (reference to the image)
					}
				}
				}
			}
		}
		$this->writer->page_end();
	}

	/* translate <> */
	function _translate(&$font, $text) {
		$map = array();
		if (isset($font->_cacheUnicodeMap)) {
			$map = $font->_cacheUnicodeMap;
		} else {
			$unicodeMap = $font->get('ToUnicode', $this->objmap);
			if($unicodeMap!==null) {
				//_debug("unicode map: $unicodeMap\n");
				//_debug("unicode map stream: {$unicodeMap->getStream()}\n");
				$streamLines = explode("\n",$unicodeMap->getStream());
				$begin = false;
				foreach($streamLines as $line) {
					if ($begin) {
						if ($line == 'endbfchar') break;
						$key = substr($line,1,2);
						$value = substr($line,6,-1);
						$map[$key] = $value;
					} else if (strstr($line, 'beginbfchar') !== FALSE) {
						$begin = true;
					}
				}
				$font->_cacheUnicodeMap = $map;
			}
		}
		//_debug( "Translate: $text");
		$r = '';
		$strlen = strlen($text);
		for($i = 0; $i<$strlen;) {
			$key = $text{$i++} . $text{$i++};
			if (array_key_exists($key, $map)) {
				$r .= chr(hexdec($map[$key]));
			}
		}
		return $r;
	}
}

function _debug($str) {
	global $debug;
	if ($debug) echo 'DEBUG: ' . $str . "\n";
}
?>
