<?php
/*
    Copyright (C) 2007 Yan Morin <yan.morin@gmail.com>

    This file is part of libfdpf.

    libfdpf is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    libfdpf is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libfdpf.  If not, see <http://www.gnu.org/licenses/>.
*/
require_once dirname(__FILE__).'/fdpf_iwriter.php';
class FDPF_SVGWriter extends FDPF_IWriter {
	var $nopage = 0;
	var $fillcolor = 'black';
	var $strokewidth = '1pt';
	var $fontfamily = '';
	var $fontsize = '';
	var $fontweight = '';
	function FPDF_SVGWriter() {}
	function _header() {
		$width = $height = '';
		if ($this->pagewidth != 0 && $this->pageheight) {
			$width = ' width="'.$this->pagewidth . $this->unit.'"';
			$height = ' height="'.$this->pageheight . $this->unit.'"';
		}
		echo '<'. '?xml version="1.0" encoding="UTF-8"?'.'>'.NL;
		echo '<svg xmlns="http://www.w3.org/2000/svg" version="1.0"'.$width.$height.'>'.NL; 
	}
	function _footer() { echo '</svg>'; }
	function page_start($o,$w,$h) { echo ' <g id="page'.$this->nopage++.'">'.NL; }
	function page_end() { echo ' </g>'.NL; }
	function text($x, $y, $txt) {
		echo '  <text x="'.$x.$this->unit.'" y="'.$y.$this->unit.'"'.$this->getstyle('text').'>'.htmlspecialchars(utf8_encode($txt)).'</text>'.NL;
	}
	function line($x1,$y1,$x2,$y2) {
		echo '  <line x1="'.$x1.$this->unit.'" y1="'.$y1.$this->unit.
		'" x2="'.$x2.$this->unit.'" y2="'.$y2.$this->unit.'" '.$this->getstyle('line').' />'.NL;
	}
	function linewidth($width) {
		$this->strokewidth = $width.$this->unit;
	}
	function rect($x,$y,$w,$h,$pdfstyle) {
		echo '  <rect x="'.$x.$this->unit.'" y="'.$y.$this->unit.
		'" width="'.$w.$this->unit.'" height="'.$h.$this->unit.'" />'.NL;
	}
	function font($family, $type, $size) {
		$this->fontfamily = $family;
		$this->fontsize = $size.$this->unit;
		$this->fontweight = $type == 'B' ? 'bold' : '';
		// todo: classify font
	}
	function fillcolor($r,$g,$b) { $this->fillcolor = 'rgb('.$r.','.$g.','.$b.')'; }
	function drawcolor($r,$g,$b) {}
	function image($id,$x,$y,$w,$h) {}
	function getstyle($obj) {
		$styleA=array();
		if ($obj == 'line') {
			if ($this->fillcolor != '')   $styleA[] = 'stroke:'.$this->fillcolor; 
			if ($this->strokewidth != '') $styleA[] = 'stroke-width:'.$this->strokewidth;
		} elseif ($obj == 'text') {
			if ($this->fontfamily != '')  $styleA[] = 'font-family:'.$this->fontfamily;
			if ($this->fontsize != '')    $styleA[] = 'font-size:'.$this->fontsize;
			if ($this->fontweight!= '')   $styleA[] = 'font-weight:'.$this->fontweight;
		}
		if (count($styleA)) {
			return ' style="'.implode(';',$styleA).'"';
		}
		return '';
	}
}

?>
